/*
 * Copyright (C) 2014 Red Hat, Inc.
 *
 * Licensed under the GNU Lesser General Public License Version 2.1
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#ifndef HY_ADVISORY_INTERNAL_H
#define HY_ADVISORY_INTERNAL_H

// libsolv
#include <solv/pool.h>

// hawkey
#include "advisory.h"

#define SOLVABLE_NAME_ADVISORY_PREFIX "patch:"

Id advisory_id(HyAdvisory advisory);
HyAdvisory advisory_create(Pool *pool, Id a_id);
int advisory_identical(HyAdvisory left, HyAdvisory right);

HyAdvisoryList advisorylist_create(Pool *pool);
void advisorylist_add(HyAdvisoryList advisorylist, HyAdvisory advisory);

#endif // HY_ADVISORY_INTERNAL_H
